if not love then error('This lib only for LOVE2D framework. download it here: https://love2d.org', 2) end
local callbacks = {'update', 'draw'}
--track love.handlers of current version of LOVE2d
for k, _ in pairs(love.handlers) do
    table.insert(callbacks, tostring(k))
end

local function isReserved(key)
	for i, v in ipairs(callbacks) do
		if key == v then return true end
	end
end

--Function returns new state like object
local function state(t)
	return setmetatable({}, {
		--table with methods
		__index = t,
		--forbidding reserved names
		__newindex = function(self, key, value)
				if not isReserved(key) then 
					rawset(self, key, value) 
				else 
					error('Trying to modify reserved callback "'..key..'"', 2) 
				end 
			end 
		}
	)
end

--State manager
local states = {}
setmetatable(states,
	{__call =
		function(self)
			local o = {}
			o.state = {}					--state list
			o.current = state()		--empty state
			o.currentname = 'nil'	--name for states:get()
			self.__index = self							--OOP
			return setmetatable(o, self)		--magic
		end
	}
)


local callbacks = {
'keypressed', 'keyreleased', 'textinput', 'textedited',
'mousepressed', 'mousereleased', 'mousefocus', 'mousemoved', 'wheelmoved', 
'touchpressed', 'touchreleased', 'touchmoved', 
'directorydropped', 'filedropped', 
'quit', 'lowmemory', 
'focus', 'visible', 
'threaderror', 'errhand',
'update', 'draw',
'gamepadaxis', 'gamepadpressed', 'gamepadreleased', 
'joystickadded', 'joystickaxis', 'joystickhat',
'joystickpressed', 'joystickreleased', 'joystickremoved'
}
--Self and additional love callback initialisation
--(load/update/draw/keypressed/keyreleased etc)
function states:init()
	for _, k in ipairs(callbacks) do
		--initialisation self methods like self:update(dt)
		self[k] = function(self, ...)
			if self.current[k] and type(self.current[k]) == 'function' then
				self.current[k](self.current, ...)
			end
		end
		--optional addition methods for love, if not defined
		if not love[k] then
			love[k] = function(...)
				self[k](self, ...)
			end
		end
	end
	return self
end

function states:new(name, t)
	self.state[name] = state(t)
end

function states:set(name, ...)
	if not name or type(name) ~= 'string' or not self.state[name] then error('Name of defined state required, got: '..tostring(name), 2) end
	if self.current.unload and type(self.current.unload) == 'function' then self.current:unload() end
	self.current = self.state[name]
	if self.current.load and type(self.current.load) == 'function' then self.current:load(...) end
	self.currentname = name
end

function states:get(name)
	if name and type(name) == 'string' and self.state[name] then return self.state[name] end
	return self.current, self.currentname
end

local s = states()
return s:init()